/********************************************************
  Xpress-BCL C Example Problems
  =============================

  file foliocb.c
  ``````````````
  Modeling a MIP problem
  to perform portfolio optimization.

  Same model as in foliomip3.c.
  -- Defining an integer solution callback --

  (c) 2009-2024 Fair Isaac Corporation
      author: S.Heipcke, May 2009, rev. Mar. 2011
********************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include "xprb.h"
#include "xprs.h"


#define MAXNUM 15                  /* Max. number of different assets */
#define MAXRISK 1.0/3              /* Max. investment into high-risk values */
#define MINREG 0.2                 /* Min. investment per geogr. region */
#define MAXREG 0.5                 /* Max. investment per geogr. region */
#define MAXSEC 0.25                /* Max. investment per ind. sector */
#define MAXVAL 0.2                 /* Max. investment per share */
#define MINVAL 0.1                 /* Min. investment per share */

#define DATAFILE "folio10.cdat"    /* File with problem data */
#define MAXENTRIES 10000


int NSHARES;                       /* Number of shares */
int NRISK;                         /* Number of high-risk shares */
int NREGIONS;                      /* Number of geographical regions */
int NTYPES;                        /* Number of share types */

double *RET;                       /* Estimated return in investment */
int *RISK;                         /* High-risk values among shares */
char **LOC;                        /* Geogr. region of shares */
char **SEC;                        /* Industry sector of shares */

char **SHARES_n;
char **REGIONS_n;
char **TYPES_n;

XPRBvar *frac;                     /* Fraction of capital used per share */
XPRBvar *buy;                      /* 1 if asset is in portfolio, 0 otherwise */

#include "readfoliodata.c_"

void XPRS_CC print_sol(XPRSprob oprob, void *vp);

int main(int argc, char **argv)
{
 int s,r,t;
 XPRBprob prob;
 XPRBctr Risk,Return,Cap,Num;
 XPRBctr *MinReg, *MaxReg, *LimSec, LinkL, LinkU;
 char *MIPSTATUS[] = {"not loaded", "not optimized", "LP optimized",
                      "unfinished (no solution)",
                      "unfinished (solution found)", "infeasible", "optimal",
                      "unbounded"};

 readdata(DATAFILE);               /* Data input from file */

 prob = XPRBnewprob("FolioMIP3");  /* Initialize a new problem in BCL */

/* Create the decision variables (including upper bounds for `frac') */
 frac = (XPRBvar*)malloc(NSHARES*sizeof(XPRBvar));
 buy = (XPRBvar*)malloc(NSHARES*sizeof(XPRBvar));
 for(s=0;s<NSHARES;s++)
 {
  frac[s] = XPRBnewvar(prob, XPRB_PL, "frac", 0, MAXVAL);
  buy[s] = XPRBnewvar(prob, XPRB_BV, "buy", 0, 1);
 }

/* Objective: total return */
 Return = XPRBnewctr(prob, "Return", XPRB_N);
 for(s=0;s<NSHARES;s++) XPRBaddterm(Return, frac[s], RET[s]);
 XPRBsetobj(prob,Return);          /* Set the objective function */

/* Limit the percentage of high-risk values */
 Risk = XPRBnewctr(prob, "Risk", XPRB_L);
 for(s=0;s<NRISK;s++) XPRBaddterm(Risk, frac[RISK[s]], 1);
 XPRBaddterm(Risk, NULL, MAXRISK);

/* Limits on geographical distribution */
 MinReg = (XPRBctr*)malloc(NREGIONS*sizeof(XPRBctr));
 MaxReg = (XPRBctr*)malloc(NREGIONS*sizeof(XPRBctr));
 for(r=0;r<NREGIONS;r++)
 {
  MinReg[r] = XPRBnewctr(prob, "MinReg", XPRB_G);
  MaxReg[r] = XPRBnewctr(prob, "MaxReg", XPRB_L);
  for(s=0;s<NSHARES;s++)
   if(LOC[r][s]>0)
   {
    XPRBaddterm(MinReg[r], frac[s], 1);
    XPRBaddterm(MaxReg[r], frac[s], 1);
   }
  XPRBaddterm(MinReg[r], NULL, MINREG);
  XPRBaddterm(MaxReg[r], NULL, MAXREG);
 }

/* Diversification across industry sectors */
 LimSec = (XPRBctr*)malloc(NTYPES*sizeof(XPRBctr));
 for(t=0;t<NTYPES;t++)
 {
  LimSec[t] = XPRBnewctr(prob, "LimSec", XPRB_L);
  for(s=0;s<NSHARES;s++)
   if(SEC[t][s]>0) XPRBaddterm(LimSec[t], frac[s], 1);
  XPRBaddterm(LimSec[t], NULL, MAXSEC);
 }

/* Spend all the capital */
 Cap = XPRBnewctr(prob, "Cap", XPRB_E);
 for(s=0;s<NSHARES;s++) XPRBaddterm(Cap, frac[s], 1);
 XPRBaddterm(Cap, NULL, 1);

/* Limit the total number of assets */
 Num = XPRBnewctr(prob, "Num", XPRB_L);
 for(s=0;s<NSHARES;s++) XPRBaddterm(Num, buy[s], 1);
 XPRBaddterm(Num, NULL, MAXNUM);

/* Linking the variables */
 for(s=0;s<NSHARES;s++)
 {
  LinkU = XPRBnewctr(prob, "LinkU", XPRB_L);
  XPRBaddterm(LinkU, frac[s], 1);
  XPRBaddterm(LinkU, buy[s], -MAXVAL);
  LinkL = XPRBnewctr(prob, "LinkL", XPRB_G);
  XPRBaddterm(LinkL, frac[s], 1);
  XPRBaddterm(LinkL, buy[s], -MINVAL);
 }

/* Define an integer solution callback */
 XPRSsetcbintsol(XPRBgetXPRSprob(prob), print_sol, prob);

/* Solve the problem */
 XPRBsetsense(prob, XPRB_MAXIM);
 XPRBmipoptimize(prob, "");

 printf("Problem status: %s\n", MIPSTATUS[XPRBgetmipstat(prob)]);

 return 0;
}

/**************************Solution printing****************************/

void XPRS_CC print_sol(XPRSprob oprob, void *vp)
{
 int num,s;
 XPRBprob bprob;

 bprob = (XPRBprob)vp;
 XPRBbegincb(bprob, oprob);

 XPRSgetintattrib(oprob, XPRS_MIPSOLS, &num); /* Get number of the solution */

 XPRBsync(bprob, XPRB_XPRS_SOL);              /* Update BCL solution values */
 printf("Solution %d: Total return: %g\n", num, XPRBgetobjval(bprob));
 for(s=0;s<NSHARES;s++)
  if(XPRBgetsol(buy[s])>0.5)
   printf("  %d : %g%% (%g)\n", s, XPRBgetsol(frac[s])*100, XPRBgetsol(buy[s]));

 XPRBendcb(bprob);
}
